//******************************************************************
//* SERVO LIN EXAMPLES 
//* WRITTEN BY FRED EADY FOR SERVO MAGAZINE
//* LAST UPDATED 05/05/2009
//* CHANGES/ADDITIONS
//******************************************************************b
#include <pic18.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <htc.h>

// Settings for PIC18F2620
__CONFIG(1,IESODIS & FCMDIS & HS);
__CONFIG(2,BORDIS & PWRTDIS & WDTDIS);
__CONFIG(3,MCLREN & LPT1DIS & PBDIGITAL);
__CONFIG(4,DEBUGEN & LVPDIS & XINSTDIS & STVREN);
__CONFIG(5,UNPROTECT);
__CONFIG(6,WRTEN);
__IDLOC(0505);
//******************************************************************
//*	FUNCTION PROTOTYPES
//******************************************************************
void putch(char c);
char recvchar(void);
char sendchar(char data);
char CharInQueue(void);
void send_lin_frame(void);
void send_lin_header(char hdr);
void load_pid (char id_lin);
void load_1(char id_lin,char data1);
void load_2(char id_lin,char data1,char data2);
void load_3(char id_lin,char data1,char data2,char data3);
void load_4(char id_lin,char data1,char data2,char data3,char data4);
void load_5(char id_lin,char data1,char data2,char data3,char data4,char data5);
void load_6(char id_lin,char data1,char data2,char data3,char data4,char data5,char data6);
void load_7(char id_lin,char data1,char data2,char data3,char data4,char data5,char data6,char data7);
void load_8(char id_lin,char data1,char data2,char data3,char data4,char data5,char data6,char data7,char data8);
//******************************************************************
//*	PIC18F2620 RAM Definitions
//******************************************************************
volatile char bytein,chksum_loc,scratch8,pid_in,id_in;
unsigned int tmsecs1,imsecs1,tsecs1,isecs1,checksum;

char lin_buf[11];
#define	msg_len	0
#define id_loc	1
#define msg_loc	2

typedef struct
{
	char enhanced: 1;
}FFLAGS;

FFLAGS flags;

typedef struct
{
	char id0: 1;
	char id1: 1;
	char id2: 1;
	char id3: 1;
	char id4: 1;
	char id5: 1;
	char p0: 1;
	char p1: 1;
}IIDBITS;

IIDBITS id_bits;

//******************************************************************
//*	PIC18F2620 ALIAS DEFINITIONS
//******************************************************************
#define	TXE		LATB0
#define	CS		LATB1
#define	LINACT	RB2
#define	LED1	LATC0
#define	OFF		0
#define ON		1
//******************************************************************
//*	MACROS
//******************************************************************
#define bitset(var, bitno) ((var) |= 1 << (bitno))
#define bitclr(var, bitno) ((var) &= ~(1 << (bitno)))
#define make8(var,offset)	((unsigned int)var >> (offset * 8)) & 0x00FF
#define	make16(varhigh,varlow)	(((unsigned int)varhigh & 0xFF)* 0x100) + ((unsigned int)varlow & 0x00FF)
#define make32(var1,var2,var3,var4) \
		((unsigned long)var1<<24)+((unsigned long)var2<<16)+ \
		((unsigned long)var3<<8)+((unsigned long)var4)		
#define make32i(var1,var2) ((unsigned long)var1<<16)+((unsigned long)var2)

#define	TX_ON		TXE = ON
#define	TX_OFF		TXE = OFF
#define	CS_ON		CS = ON
//#define CS_OFF		CS = OFF
#define	LIN_IDLE	CS_ON;	\
					TX_OFF;
#define	LIN_RX		CS_ON;	\
					TX_OFF;	\
					CREN = 0;	\
					CREN = 1;
#define LIN_TX		CS_ON;	\
					TX_ON;	\
					CREN = 0;
				
#define SET_WUE		WUE = 1
#define SET_ABDEN	ABDEN = 1
//******************************************************************
//*	EUSART BUFFER DEFINITIONS
//******************************************************************
#define BAUD	19200		//desired baud rate
#define FOSC	20000000	//oscillator frequency
#define DIVIDER ((unsigned int)(FOSC/(16 * BAUD) - 1))
//1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_RX_BUFFER_SIZE  256     
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )
// 1,2,4,8,16,32,64,128 or 256 bytes 
#define USART_TX_BUFFER_SIZE  256      
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )
char USART_RxBuf[USART_RX_BUFFER_SIZE],USART_TxBuf[USART_TX_BUFFER_SIZE];
char USART_TxHead,USART_TxTail,USART_RxHead,USART_RxTail;
//******************************************************************
//*	INTERRUPT DEFINITIONS
//******************************************************************
#define		disable_RXint	  	RCIE = 0
#define		enable_RXint	  	RCIE = 1
#define		disable_TMR1int  	TMR1IE = 0
#define		enable_TMR1int	 	TMR1IE = 1

#define		enable_GLOBALint  	IPEN = 0;				\
							  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	IPEN = 0;				\
							  	GIE = 0;				\
							  	PEIE = 0;
							  								  	
#define		TIMER1ON			TMR1ON = 1
#define		TIMER1OFF			TMR1ON = 0

#define mdelay1(msecdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								TIMER1ON;					\
								while(tmsecs1 < msecdelay);
								
#define	sdelay1(secdelay)		TIMER1OFF;					\
								TMR1IF = 0;					\
								imsecs1 = 0;				\
								tmsecs1 = 0;				\
								isecs1 = 0;					\
								tsecs1 = 0;					\
								TIMER1ON;					\
								while(tsecs1 < secdelay);
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}

#define esc	0x1B
//******************************************************************
//*	INTERRUPT HANDLER ROUTINE
//******************************************************************
void interrupt TIMERS(void)
{
  char usart_data,tmphead,tmptail;
  
	if((TMR1IF && TMR1IE))
	{
		TMR1IF = 0;
		TMR1H = 0xEC;	
		TMR1L = 0x77;

		++tmsecs1;
		if(++imsecs1 == 1000)
		{
			imsecs1 = 0;
			++tsecs1;
			++isecs1;
		}
	}

  if(RCIF && RCIE)
 {
   usart_data = RCREG;                   // read the received usart_data 
                                // calculate buffer index 
   tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK;
   USART_RxHead = tmphead;        // store new index 

   if ( tmphead == USART_RxTail )
   {
     // ERROR! Receive buffer overflow 
   }
                                
   USART_RxBuf[tmphead] = usart_data;   // store received usart_data in buffer 
  }

  
  if(TRMT && TXIE)
  {
                                // check if all usart_data is transmitted  
   if ( USART_TxHead != USART_TxTail )
   {
                                // calculate buffer index  
     tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
     USART_TxTail = tmptail;      // store new index  

    TXREG = USART_TxBuf[tmptail];  // start transmission
   }
   else
   {
     TXIE = 0;         // disable TX interrupt  
   }
  }
} 
//******************************************************************
//*	Init EUSART Function
//******************************************************************
void init_EUSART(void)
{
// load baud rate divisor for 19200 bps with 20MHz clock  = 64
// load baud rate divisor for 9600 bps with 20MHz clock  = 129
  //SPBRG = 129;	//uncomment for master node operation	
  TRISC7 = 1;		//receive pin
  TRISC6 = 0;		//transmit pin
  TXSTA = 0x04;		//high speed baud rate	
  RCSTA = 0x80;		//enable serial port and serial port pins
  
  USART_RxTail = 0x00;	//flush receive buffer 
  USART_RxHead = 0x00;
  USART_TxTail = 0x00;	//flush transmit buffer
  USART_TxHead = 0x00;
  RCIP = 1;			//receive interrupt = high priority
  TXIP = 1;			//transmit interrupt = high priority
  RCIE = 1;			//enable receive interrupt
  PEIE = 1;			//enable all unmasked peripheral interrupts
  GIE = 1;			//enable all unmasked interrupts
  CREN = 1;			//enable USART1 receiver
  TXIE = 0;			//disable USART1 transmit interrupt
  TXEN = 1;			//transmitter enabled
}
//****************************************************************************
//*	INITIALIZE ROUTINE
//****************************************************************************
void init(void)
{
   	TRISA = 0b11111111;		
   	TRISB = 0b11111100;		
   	TRISC = 0b11111110;
//******************************************************************
//*	CONFIGURE MCP2021
//******************************************************************
   	LIN_IDLE;
//******************************************************************
//*	CONFIGURE EEPROM READ/WRITE
//******************************************************************
	EECON1 = 0b00000000;
//******************************************************************
//*	CONFIGURE AND START TIMER1
//* SET TO OVERFLOW EVERY 1mS
//******************************************************************
	TIMER1OFF;
	T1CON = 0b00000000;
	TMR1H = 0xEC;
	TMR1L = 0x77;
	TIMER1ON;
//******************************************************************
//*	CONFIGURE A2D AND COMPARATORS
//******************************************************************
	ADCON0 = 0b00000000;	//turn OFF A2D
	ADCON1 = 0b00001111;	//all digital
//******************************************************************
//*	INITIALIZE THE EUSART
//******************************************************************
	init_EUSART();
//******************************************************************
//*	GENERAL HOUSEKEEPING
//******************************************************************
	LED1 = OFF;
	flags.enhanced = 1;
//******************************************************************
//*	ACTIVATE INTERRUPTS
//******************************************************************
	enable_TMR1int;
    enable_GLOBALint;
//******************************************************************
}
//******************************************************************
//*	GET A CHARACTER FROM RECEIVE BUFFER
//******************************************************************
char recvchar(void)
{
  char tmptail;
                                // wait for incomming usart_data  
  while ( USART_RxHead == USART_RxTail );
                                // calculate buffer index  
  tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK;
  USART_RxTail = tmptail;        // store new index  

  return USART_RxBuf[tmptail];   // return usart_data  
}

//******************************************************************
//*	SEND A CHARACTER
//******************************************************************
char sendchar(char usart_data)
{
  char tmphead;
                                // calculate buffer index 
  tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK;
                                // wait for free space in buffer 
  while ( tmphead == USART_TxTail );
                                // store usart_data in buffer 
  USART_TxBuf[tmphead] = usart_data;
  USART_TxHead = tmphead;        // store new index 

  TXIE = 1;            // enable TX interrupt 

  return usart_data;
}
//******************************************************************
//*	CHECK FOR CHARACTER IN RECEIVE BUFFER
//******************************************************************
char CharInQueue(void)
{
  return(USART_RxHead != USART_RxTail);
}
//******************************************************************
//*	SEND A LIN FRAME
//******************************************************************
void send_lin_frame(void)
{
	char i,id_loc_org;
	
	id_loc_org = lin_buf[id_loc];
	//CALCULATE PROTECTED ID CHECKSUM
	id_bits.id0 =  lin_buf[id_loc] & 0b00000001;
	id_bits.id1 = (lin_buf[id_loc] & 0b00000010) >> 1;
	id_bits.id2 = (lin_buf[id_loc] & 0b00000100) >> 2;
	id_bits.id3 = (lin_buf[id_loc] & 0b00001000) >> 3;
	id_bits.id4 = (lin_buf[id_loc] & 0b00010000) >> 4;
	id_bits.id5 = (lin_buf[id_loc] & 0b00100000) >> 5;
	id_bits.p0 = id_bits.id0 ^ id_bits.id1 ^ id_bits.id2 ^ id_bits.id4;
	id_bits.p1 = ~(id_bits.id1 ^ id_bits.id3 ^ id_bits.id4 ^ id_bits.id5);
	if(id_bits.p1)
	{
		lin_buf[id_loc] = lin_buf[id_loc] + 0b10000000;
	}
	if(id_bits.p0)
	{
		lin_buf[id_loc] = lin_buf[id_loc] + 0b01000000;
	}
	//CALCULATE FRAME CHECKSUM
	if(id_loc_org >= 0x3C && id_loc_org <= 0x3F)
	{
  		checksum = 0;				//for classic checksum calculation
 	}	 	
	else if(flags.enhanced)
	{
  		checksum = lin_buf[id_loc];	//for enhanced checksum calculation
 	}
	for(i=0;i<lin_buf[msg_len];++i)
	{	
 		checksum = checksum + lin_buf[msg_loc+i];
 	}
  	lin_buf[chksum_loc] = ~(checksum & 0x00FF);

  	lin_buf[msg_len] = lin_buf[msg_len] + 2; //id and checksum added to length
  	LIN_TX;			//place MCP2201 in transmit mode
  	SENDB = 1;		//set SEND BREAK bit
  	sendchar(0x55);	//send dummy character
  	sendchar(0x55);	//send autobaud chirp
  	//SEND FRAME
  	for(i=0;i<lin_buf[msg_len];++i)
  	{
		sendchar(lin_buf[id_loc+i]);
	}
	mdelay1(17);
}
//******************************************************************
//*	SEND A LIN HEADER
//******************************************************************
void send_lin_header(char hdr)
{
	char i;
	
 	lin_buf[id_loc] = hdr;

	//CALCULATE PROTECTED ID CHECKSUM
	id_bits.id0 =  lin_buf[id_loc] & 0b00000001;
	id_bits.id1 = (lin_buf[id_loc] & 0b00000010) >> 1;
	id_bits.id2 = (lin_buf[id_loc] & 0b00000100) >> 2;
	id_bits.id3 = (lin_buf[id_loc] & 0b00001000) >> 3;
	id_bits.id4 = (lin_buf[id_loc] & 0b00010000) >> 4;
	id_bits.id5 = (lin_buf[id_loc] & 0b00100000) >> 5;
	id_bits.p0 = id_bits.id0 ^ id_bits.id1 ^ id_bits.id2 ^ id_bits.id4;
	id_bits.p1 = ~(id_bits.id1 ^ id_bits.id3 ^ id_bits.id4 ^ id_bits.id5);
	if(id_bits.p1)
	{
		lin_buf[id_loc] = lin_buf[id_loc] + 0b10000000;
	}
	if(id_bits.p0)
	{
		lin_buf[id_loc] = lin_buf[id_loc] + 0b01000000;
	}

  	LIN_TX;			//place MCP2201 in transmit mode
  	SENDB = 1;		//set SEND BREAK bit
  	sendchar(0x55);	//send dummy character
  	sendchar(0x55);	//send autobaud chirp
	sendchar(lin_buf[id_loc]);
	LIN_RX;
	mdelay1(17);
}
void load_pid (char id_lin)
{	
 	lin_buf[msg_len] = 0;
 	lin_buf[id_loc] = id_lin;
 	chksum_loc = msg_loc;
}
void load_1(char id_lin,char data1)
{	
 	lin_buf[msg_len] = 1;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	chksum_loc = msg_loc + 1;
}
void load_2(char id_lin,char data1,char data2)
{	
 	lin_buf[msg_len] = 2;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	chksum_loc = msg_loc + 2;
}
void load_3(char id_lin,char data1,char data2,char data3)
{	
 	lin_buf[msg_len] = 3;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	lin_buf[msg_loc+2] = data3;
 	chksum_loc = msg_loc + 3;
}
void load_4(char id_lin,char data1,char data2,char data3,char data4)
{	
 	lin_buf[msg_len] = 4;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	lin_buf[msg_loc+2] = data3;
 	lin_buf[msg_loc+3] = data4;
 	chksum_loc = msg_loc + 4;
}
void load_5(char id_lin,char data1,char data2,char data3,char data4,char data5)
{	
 	lin_buf[msg_len] = 5;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	lin_buf[msg_loc+2] = data3;
 	lin_buf[msg_loc+3] = data4;
 	lin_buf[msg_loc+4] = data5;
 	chksum_loc = msg_loc + 5;
}
void load_6(char id_lin,char data1,char data2,char data3,char data4,char data5,char data6)
{	
 	lin_buf[msg_len] = 6;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	lin_buf[msg_loc+2] = data3;
 	lin_buf[msg_loc+3] = data4;
 	lin_buf[msg_loc+4] = data5;
 	lin_buf[msg_loc+5] = data6;
 	chksum_loc = msg_loc + 6;
}
void load_7(char id_lin,char data1,char data2,char data3,char data4,char data5,char data6,char data7)
{	
 	lin_buf[msg_len] = 7;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	lin_buf[msg_loc+2] = data3;
 	lin_buf[msg_loc+3] = data4;
 	lin_buf[msg_loc+4] = data5;
 	lin_buf[msg_loc+5] = data6;
 	lin_buf[msg_loc+6] = data7;
 	chksum_loc = msg_loc + 7;
}
void load_8(char id_lin,char data1,char data2,char data3,char data4,char data5,char data6,char data7,char data8)
{	
 	lin_buf[msg_len] = 8;
 	lin_buf[id_loc] = id_lin;
 	lin_buf[msg_loc] = data1;
 	lin_buf[msg_loc+1] = data2;
 	lin_buf[msg_loc+2] = data3;
 	lin_buf[msg_loc+3] = data4;
 	lin_buf[msg_loc+4] = data5;
 	lin_buf[msg_loc+5] = data6;
 	lin_buf[msg_loc+6] = data7;
 	lin_buf[msg_loc+7] = data8;
 	chksum_loc = msg_loc + 8;
}
//******************************************************************
//*	MAIN FUNCTION
//******************************************************************
void main(void)
{
	char i;
  	init();				//init the EUSART
/*
	//SEND HEADER EXAMPLE  	
  	send_lin_header(0x01);
  	do{
	  	++scratch8;
	  }while(1);
*/

/*	
	//SEND A FULL FRAME EXAMPLE
  	do{
  	load_8(0x3C,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08);
  	send_lin_frame();
  	load_7(0x3D,0x01,0x02,0x03,0x04,0x05,0x06,0x07);
  	send_lin_frame();
	load_6(0x3E,0x01,0x02,0x03,0x04,0x05,0x06);
  	send_lin_frame();
  	load_5(0x3F,0x01,0x02,0x03,0x04,0x05);
  	send_lin_frame();
  	load_4(0x04,0x01,0x02,0x03,0x04);
  	send_lin_frame();
  	load_3(0x05,0x01,0x02,0x03);
  	send_lin_frame();
  	load_2(0x06,0x01,0x02);
  	send_lin_frame();
  	load_1(0x07,0x01);
  	send_lin_frame();
  	load_pid(0x01);
  	send_lin_frame();
  	}while(1);
*/

/*
	//SLAVE RECEIVER BREAK-SYNC EXAMPLE
  	if(RCIDL)			//make sure receive not in progress
  	{
	  	LIN_RX;			//put MCP2021 in receive mode
  		disable_RXint;	//disable the PIC's rx interrupt
  		SET_WUE;		//enable PIC's WakeUpEnable feature
  		SET_ABDEN;		//enable PIC's autobaud feature
  		IDLEN = 1;		//idle CPU only while sleeping
  		SLEEP();		//snooze
  		while(WUE);		//wait for Break sequence to complete
     	bytein = RCREG;	//discard the byte-clear RCIF flag                   
  		while(ABDEN);	//measure the autobaud chirp
     	bytein = RCREG;	//discard the byte- clear RCIF flag                   
		enable_RXint;	//get ready for real data
	}
  	do{
	  	LED1 ^= 1;
	  	sdelay1(2);
	  }while(1);	
*/ 

/*	
	//SLAVE RECEIVER WITH RESPONSE EXAMPLE
  	if(RCIDL)			//make sure receive not in progress
  	{
	  	LIN_RX;			//put MCP2021 in receive mode
  		disable_RXint;	//disable the PIC's rx interrupt
  		SET_WUE;		//enable PIC's WakeUpEnable feature
  		SET_ABDEN;		//enable PIC's autobaud feature
  		IDLEN = 1;		//idle CPU only while sleeping
  		SLEEP();		//snooze
  		while(WUE);		//wait for Break sequence to complete
     	bytein = RCREG;	//discard the byte-clear RCIF flag                   
  		while(ABDEN);	//measure the autobaud chirp
     	bytein = RCREG;	//discard the byte- clear RCIF flag
		enable_RXint;	//get ready for real data
		while(!(CharInQueue()));	//wait for incoming PID
		pid_in = recvchar();		//retrieve PID from receive buffer ring
		id_in = pid_in & 0x3F;		//get identifier
		if(id_in == 0x02)
		{	
			lin_buf[msg_len] = 3;		//number of data bytes in response
			lin_buf[id_loc] = pid_in;	//used to compute frame checksum
			lin_buf[msg_loc] = 0x11;	//response data
			lin_buf[msg_loc+1] = 0x22;	//response data
			lin_buf[msg_loc+2] = 0x33;	//response data
	  		checksum = lin_buf[id_loc];	//enhanced checksum calculation
			for(i=0;i<lin_buf[msg_len];++i)
			{	
		 		checksum = checksum + lin_buf[msg_loc+i];
		 	}
		 	chksum_loc = msg_loc + 3;	//where to store frame checksum
		  	lin_buf[chksum_loc] = ~(checksum & 0x00FF); //frame checksum
		
		  	lin_buf[msg_len] = lin_buf[msg_len] + 1; //checksum added to length
		  	LIN_TX;			//place MCP2201 in transmit mode
		  	//SEND RESPONSE
		  	for(i=0;i<lin_buf[msg_len];++i)
		  	{
				sendchar(lin_buf[msg_loc+i]);
			}
		  	do{				//kill time
			  	LED1 ^= 1;
			  	sdelay1(2);
			  }while(1);
		}	
	}
	 
*/
	 
}
